---
title: ebnf-parse
section: 1
---

## NAME

ebnf-parse - validate and apply EBNF grammars

## SYNOPSIS

```
ebnf-parse [OPTIONS] EBNF_FILE [INPUT_FILE]
```

Options:

```
  -h               --help                Display help and quit
  -v               --version             Display version and quit
  -p PREDEF_FILE   --predef=PREDEF_FILE  Specify a predefined-terms file
  -s START_SYMBOL  --start=START_SYMBOL  Specify the start symbol
  -a               --allow-spaces        Allow optional spaces after each terminal
```

## DESCRIPTION

`ebnf-parse` is a command-line tool that validates and processes grammars written in
Extended Backus-Naur Form (EBNF), specifically using the syntax accepted by the [BNF
Playground](https://bnfplayground.pauliankline.com). It checks whether the provided EBNF
grammar file (`EBNF_FILE`) is valid, and, optionally, whether it's possible to parse an
input file (`INPUT_FILE`) according to that grammar.

### Predefined Non-Terminals

The tool allows you to use **predefined non-terminals**, which are defined externally in a
separate file. These predefined terms act as "stubs" – representing parts of the grammar
that are not fully implemented but are required for parsing. For example, you could define a
non-terminal such as `<number>` that always evaluates to `"42"` in the predefined file, while
the actual implementation may differ in a production system.

You can provide this file using the `-p` or `--predef` option, specifying the file path of
the predefined terms.

### Parsing an Input File

If you provide an `INPUT_FILE` along with the `EBNF_FILE`, `ebnf-parse` will attempt to
parse the input using the grammar you defined. In this scenario, you must also specify a
**start symbol** with the `-s` or `--start` option, which indicates the top-level
non-terminal to begin parsing from. The start symbol typically corresponds to the main
concept your input represents, like `<expression>`, `<program>`, or in the example shown
below under **EXAMPLES**, `<gpa>`.

### Allowing Optional Spaces

When parsing an input file, you can enable the `-a` or `--allow-spaces` option.  This
tells `ebnf-parse` to automatically treat every terminal in the grammar as optionally
followed by a space (`" "?`). This is particularly useful for languages where tokens or
words are typically separated by whitespace but the exact spacing might vary in the input.


### Output

If the input file is valid according to the grammar, `ebnf-parse` will display a [parse
tree](https://en.wikipedia.org/wiki/Parse_tree) for the input and exit normally. If the
input or grammar is invalid, it will output an error message and return a non-zero exit
status.

### EBNF Dialect

There are many variants of Extended Backus-Naur Form (EBNF). `ebnf-parse`, like the BNF
Playground, uses a variant of EBNF similar to the notation typically used for **regular
expressions** (see <https://en.wikipedia.org/wiki/Regular_expression#Basic_concepts>).
This is one of the most common and intuitive EBNF variants.

Another EBNF variant was formalized by Niklaus Wirth, and uses curly braces (`{}`) to
represent repetition. `ebnf-parse` does not support this syntax.

Additionally, `ebnf-parse` does not support the *augmented* BNF (ABNF) variant, which uses
square brackets (`[]`) for optional items and a leading asterisk (`*`) to denote
repetition. Although ABNF is a very comprehensive notation, it is less intuitive than
other variants for most programmers, as well as being very ugly.

For a more detailed discussion on the various EBNF dialects and their trade-offs, see Pete
Jinks' web-page, "Notations for Context-Free Grammars," available at:
<http://www.cs.man.ac.uk/~pjj/bnf/bnf.html>.

## OPTIONS

`-h, --help`

:   Display help information and exit.
  
`-v, --version`

:   Display the current version of `ebnf-parse` and exit.
  
`-p PREDEF_FILE, --predef=PREDEF_FILE`

:   Specify a file containing predefined non-terminals. This file provides stubbed
    definitions for non-terminals referenced in the grammar but not fully defined.

`-s START_SYMBOL, --start=START_SYMBOL`

:   Specify the start symbol (without angle brackets) for parsing the input file. A start
    symbol is needed when parsing an input file; if one is not provided explicitly on
    the command-line, it will default to `main`.

`-a, --allow-spaces`

:   Allow optional spaces after each terminal when parsing the input file.

## EXAMPLES

### Example 1: Simple GPA Grammar

Suppose we have a simple EBNF grammar for parsing student GPAs, stored in a file
`gpa.ebnf`:

```
<gpa> ::= "4.0" | <leading> "." <trailing> | <special>
<leading> ::= [0-3]
<trailing> ::= [0-9]
```

In this grammar, `<gpa>` represents a GPA, which can either be `"4.0"`, a number like
`"3.2"`, or a special case handled by `<special>`. Since we haven't defined `<special>`,
we can stub it out using a predefined terms file `predef.ebnf`:

```
<special> ::= "TBA"
```

To parse an example input file `sample_gpa.txt`, containing the GPA `"3.2"`, we can run:

```bash
ebnf-parse -s gpa -p ./predef.ebnf ./gpa.ebnf ./sample_gpa.txt
```

If the input matches the grammar, the tool will display a parse tree for the input and
exit normally. If the input is invalid, or if there is an issue with the grammar, it will
output an error message and exit with a non-zero status.

### Example 2: Parsing with Optional Spaces

If the input file might have optional spaces between tokens, you can use the `-a` option.
For example, if `sample_gpa.txt` contains `"3 . 2"` (with spaces), you would run:

```bash
ebnf-parse -s gpa -p ./predef.ebnf -a ./gpa.ebnf ./sample_gpa.txt
```

This allows for flexible input formatting while still adhering to the grammar.

### PORTABILITY

`ebnf-parse` is implemented in the platform-independent functional programming language
[Haskell](https://www.haskell.org), using the [`ghc`](https://www.haskell.org/ghc/)
compiler. Thanks to Haskell’s cross-platform nature, it should be straightforward to
compile and run `ebnf-parse` on a wide range of common operating systems, including macOS
and Windows. However, at present, only a Linux binary is provided for direct download.
Users on other platforms will need to compile the tool from source.


## LIMITATIONS

`ebnf-parse` is capable of handling large EBNF grammars with hundreds of non-terminals
efficiently. On modern systems, it can process such grammars in a fraction of a second,
using only a moderate amount of memory (typically a few megabytes). This makes it more
memory-efficient than the BNF Playground, which can handle only grammars of around a
dozen non-terminals before running out of memory.

Despite its ability to parse recursive grammars, `ebnf-parse` currently cannot
*apply* these grammars to input files if recursion is involved. Attempting to do so will
cause the program to enter an infinite loop and hang. This limitation arises from
`ebnf-parse`'s use of a **recursive descent parser** with lookahead.
In contrast, the BNF Playground employs the [Earley
algorithm](https://en.wikipedia.org/wiki/Earley_parser), which can handle any context-free
grammar, but at the cost of potentially higher memory usage.

Future updates to `ebnf-parse` aim to address this limitation by retaining the recursive
descent approach for parsing EBNF grammars, while switching to the Earley algorithm when
parsing input files, thereby allowing efficient handling of recursive grammars without the
risk of infinite loops.

## SEE ALSO

For more information on EBNF and parse trees, see:

- [BNF Playground](https://bnfplayground.pauliankline.com)
- [Parse Tree (Wikipedia)](https://en.wikipedia.org/wiki/Parse_tree)

## AUTHOR

Arran D. Stewart

<!--
  vim: tw=90 :
-->
